@@ -4,16 +4,17 @@
   https://github.com/trptcolin/reply/"
   {:author "Chas Emerick"}
   (:require
-   [clojure.java.io :as io]
    [clojure.edn :as edn]
+   [clojure.java.io :as io]
    [clojure.string :as str]
+   [nrepl.ack :refer [send-ack]]
    [nrepl.config :as config]
    [nrepl.core :as nrepl]
-   [nrepl.ack :refer [send-ack]]
-   [nrepl.misc :refer [noisy-future]]
+   [nrepl.misc :refer [uuid]]
    [nrepl.server :as nrepl-server]
    [nrepl.socket :as socket]
    [nrepl.transport :as transport]
+   [nrepl.util.threading :as threading]
    [nrepl.version :as version])
   (:import
    [java.net URI]))
@@ -93,13 +94,23 @@ Exit:      Control+D or (exit) or (quit)"
                      err    print
                      out    print
                      value  println}}]
-  (let [client (nrepl/client transport Long/MAX_VALUE)]
+  (let [client (nrepl/client transport Long/MAX_VALUE)
+        awaiting (atom {})]
     (println (repl-intro))
+    (threading/run-with @threading/transport-executor
+      ;; Doesn't matter which executor we hijack here, just don't use `future`.
+      (doseq [res (client)]
+        (some-> (:out res) out)
+        (some-> (:err res) err)
+        (when-let [p (and (some #{:done "done"} (:status res))
+                          (get @awaiting (:id res)))]
+          (deliver p true))
+        (when (some-> ^String (:id res)
+                      (.startsWith "nrepl.cmdline-"))
+          (some-> (:value res) value))
+        (flush)))
     ;; We take 50ms to listen to any greeting messages, and display the value
     ;; in the `:out` slot.
-    (noisy-future (->> (client)
-                       (take-while #(nil? (:id %)))
-                       (run! #(when-let [msg (:out %)] (print msg)))))
     (Thread/sleep 50)
     (let [session (nrepl/client-session client)]
       (swap! running-repl assoc :transport transport)
@@ -111,24 +122,29 @@ Exit:      Control+D or (exit) or (quit)"
           (let [input (read *in* false 'exit)]
             (if (done? input)
               (clean-up-and-exit 0)
-              (do (doseq [res (nrepl/message session {:op "eval" :code (pr-str input)})]
-                    (when (:value res) (value (:value res)))
-                    (when (:out res) (out (:out res)))
-                    (when (:err res) (err (:err res)))
-                    (when (:ns res) (set! *ns* (create-ns (symbol (:ns res))))))
-                  (recur)))))))))
+              ;; Make sure the metadata read from *in* is preserved when we feed
+              ;; the form to nREPL.
+              (let [code-str (binding [*print-meta* true]
+                               (pr-str input))
+                    id (str "nrepl.cmdline-" (uuid))]
+                (swap! awaiting assoc id (promise))
+                (doseq [res (nrepl/message session {:op "eval" :code code-str :id id})]
+                  (when (:ns res) (set! *ns* (create-ns (symbol (:ns res))))))
+                @(get @awaiting id)
+                (swap! awaiting dissoc id)
+                (recur)))))))))
 
 (defn- run-repl
   ([{:keys [server options]}]
    (let [{:keys [host port socket] :or {host "127.0.0.1"}} server
-         {:keys [transport] :or {transport #'transport/bencode}} options]
+         {:keys [transport tls-keys-file tls-keys-str] :or {transport #'transport/bencode}} options]
      (run-repl-with-transport
       (cond
         socket
         (nrepl/connect :socket socket :transport-fn transport)
 
         (and host port)
-        (nrepl/connect :host host :port port :transport-fn transport)
+        (nrepl/connect :host host :port port :transport-fn transport :tls-keys-file tls-keys-file :tls-keys-str tls-keys-str)
 
         :else
         (die "Must supply host/port or socket."))
@@ -191,7 +207,8 @@ Exit:      Control+D or (exit) or (quit)"
 
 (defn help
   []
-  (str "Usage:
+  ;; Updating this? Remember to also update server.adoc
+  "Usage:
 
   -i/--interactive            Start nREPL and connect to it with the built-in client.
   -c/--connect                Connect to a running nREPL with the built-in client.
@@ -201,12 +218,12 @@ Exit:      Control+D or (exit) or (quit)"
   -p/--port PORT              Start nREPL on PORT. Defaults to 0 (random port) if not specified.
   -s/--socket PATH            Start nREPL on filesystem socket at PATH or nREPL to connect to when using --connect.
   --ack ACK-PORT              Acknowledge the port of this server to another nREPL server running on ACK-PORT.
-  -n/--handler HANDLER        The nREPL message handler to use for each incoming connection; defaults to the result of `(nrepl.server/default-handler)`.
-  -m/--middleware MIDDLEWARE  A sequence of vars, representing middleware you wish to mix in to the nREPL handler.
-  -t/--transport TRANSPORT    The transport to use. By default that's nrepl.transport/bencode.
+  -n/--handler HANDLER        The nREPL message handler to use for each incoming connection; defaults to the result of `(nrepl.server/default-handler)`. Must be expressed as a namespace-qualified symbol. The underlying var will be automatically `require`d.
+  -m/--middleware MIDDLEWARE  A sequence of vars (expressed as namespace-qualified symbols), representing middleware you wish to mix in to the nREPL handler. The underlying vars will be automatically `require`d.
+  -t/--transport TRANSPORT    The transport to use (default `nrepl.transport/bencode`), expressed as a namespace-qualified symbol. The underlying var will be automatically `require`d.
   --help                      Show this help message.
   -v/--version                Display the nREPL version.
-  --verbose                   Show verbose output."))
+  --verbose                   Show verbose output.")
 
 (defn- require-and-resolve
   "Attempts to resolve the config `key`'s `value` as a namespaced symbol
@@ -362,14 +379,16 @@ Exit:      Control+D or (exit) or (quit)"
    :host (:host options)
    :socket (:socket options)
    :transport (options->transport options)
-   :repl-fn (options->repl-fn options)})
+   :repl-fn (options->repl-fn options)
+   :tls-keys-str (:tls-keys-str options)
+   :tls-keys-file (:tls-keys-file options)})
 
 (defn server-opts
   "Takes a map of nREPL CLI options
   Returns map of processed options to start an nREPL server."
   [options]
   (let [middleware (sanitize-middleware-option (:middleware options))
-        {:keys [host port socket transport]} (connection-opts options)]
+        {:keys [host port socket transport tls-keys-str tls-keys-file]} (connection-opts options)]
     (merge options
            {:host host
             :port port
@@ -377,6 +396,8 @@ Exit:      Control+D or (exit) or (quit)"
             :transport transport
             :bind (:bind options)
             :middleware middleware
+            :tls-keys-str tls-keys-str
+            :tls-keys-file tls-keys-file
             :handler (options->handler options middleware)
             :greeting (options->greeting options transport)
             :ack-port (options->ack-port options)
@@ -401,7 +422,8 @@ Exit:      Control+D or (exit) or (quit)"
                                 {:transport transport})})
       (repl-fn host port
                (merge (when (:color options) colored-output)
-                      {:transport transport})))))
+                      {:transport transport}
+                      (select-keys options [:tls-keys-str :tls-keys-file]))))))
 
 (defn connect-to-server
   "Connects to a running nREPL server and runs a REPL. Exits program when REPL
@@ -461,14 +483,16 @@ Exit:      Control+D or (exit) or (quit)"
   "Creates an nREPL server instance.
   Takes map of CLI options.
   Returns nREPL server map."
-  [{:keys [port bind socket handler transport greeting]}]
+  [{:keys [port bind socket handler transport greeting tls-keys-str tls-keys-file]}]
   (nrepl-server/start-server
    :port port
    :bind bind
    :socket socket
    :handler handler
    :transport-fn transport
-   :greeting-fn greeting))
+   :greeting-fn greeting
+   :tls-keys-str tls-keys-str
+   :tls-keys-file tls-keys-file))
 
 (defn dispatch-commands
   "Look at options to dispatch a specified command.
